local Icon = LibStub("LibDBIcon-1.0")
local AceConfig = LibStub("AceConfig-3.0")
local AceDB = LibStub("AceDB-3.0")
local Dialog = LibStub("LibDialog-1.0")

local LDBobject = LibStub("LibDataBroker-1.1"):NewDataObject("BossList", {
	type = "launcher",
	icon = "Interface\\Buttons\\UI-GroupLoot-Dice-Up",
  OnTooltipShow = function(tooltip) 
    tooltip:AddLine("|cFFFFFFFFBossList|r")
    tooltip:AddLine("Click to open the overview of needs")
    tooltip:AddLine("Right click to open the settings") end,
  label = "BossList",
	OnClick = function() 
    if GetMouseButtonClicked() == "RightButton" then
      InterfaceOptionsFrame_OpenToCategory("BossList");
    else
      PlaySound("igMainMenuOptionCheckBoxOn")
      BossList.frame:Show()
    end end,
})


local DB

BossList.REVISION = 3

local frame = CreateFrame("Frame")
local DisplayEncounter_Orig
local dropDown

local desireNames = {"BiS", "Upgrade", "Off-spec", "Not needed"}
local desireIcons = {"B", "U", "O", " "}
local desireColors = {{r=0, g=1, b=0}, {r=1, g=1, b=0}, {r=1, g=0.5, b=0}, {r=0, g=0, b=0}}
local difficultyNames = {"Normal", "Heroic", "Flex"}

local playerName = GetUnitName("player", false)

local queriedItems = {}

local difficulty = 0


local defaults = {
  char = {
    minimap = {
      hide = false,
    }
  },
  global = {
    hideNotNeeded = true,
    blackList = {},
    revision = BossList.REVISION,
  }
}


function BossList_DEBUG_PrintEntriesIDs(table)
  print("Printing contents of table")
  local table = table or BossList_entries
  print(table)
  for instance, difficulties in pairs(table) do               -- Instances
    for difficulty, encounters in pairs(difficulties) do      -- Difficulties
      for encounter, items in pairs(encounters) do            -- Encounters
        for item, desires in pairs(items) do                  -- Items
          for desire, characters in pairs(desires) do         -- Desires
            for character, lastUpdate in pairs(characters) do -- Characters
              local msg = string.format("%d-%d-%d-%d-%s-%d-%d", instance, difficulty, encounter, item, character, desire, lastUpdate)
              print(msg)
            end
          end
        end
      end
    end
  end
end

function BossList_DEBUG_PrintEntries(table)
  print("Printing contents of table")
  local table = table or BossList_entries
  print(table)
  for instance, difficulties in pairs(table) do               -- Instances
    local instanceName = EJ_GetInstanceInfo(instance) or "Unknown"
    for difficulty, encounters in pairs(difficulties) do      -- Difficulties
      for encounter, items in pairs(encounters) do            -- Encounters
        local encounterName = EJ_GetEncounterInfo(encounter) or "Unknown"
        for item, desires in pairs(items) do                  -- Items
          local itemName = GetItemInfo(item)
          for desire, characters in pairs(desires) do         -- Desires
            for character, lastUpdate in pairs(characters) do -- Characters
              lastUpdateTime = date("%d/%m/%y %H:%M:%S", lastUpdate) or "Unknown"
              local msg = string.format("%s-%d-%s-%s-%s-%d-%s", instanceName, difficulty, encounterName, itemName, character, desire, lastUpdateTime)
              print(msg)
            end
          end
        end
      end
    end
  end
end

function BossList.QueryItem(itemID, textField)
  if queriedItems[itemID] == nil then
    queriedItems[itemID] = {}
  end
  
  tinsert(queriedItems[itemID], textField)
end

-- Return the entry if it exists, otherwise return nil
local function getEntry(table, instanceID, difficulty, encounterID, itemID, characterName)
  if table[instanceID] == nil or table[instanceID][difficulty] == nil or  -- Check if the encounter table exists
     table[instanceID][difficulty][encounterID] == nil or table[instanceID][difficulty][encounterID][itemID] == nil then

    return nil
  end
  
  -- Check all desires for an entry
  for desire, characters in pairs(table[instanceID][difficulty][encounterID][itemID]) do
    if characters[characterName] ~= nil then
      return desire, characters[characterName]
    end
  end
  
  -- Nothing found
  return nil
end


-- Add entry into table
local function addEntry(table, instanceID, difficulty, encounterID, itemID, desirability, characterName, lastUpdate)
  if table[instanceID] == nil then table[instanceID] = {} end
  local instanceTable = table[instanceID]
  if instanceTable[difficulty] == nil then instanceTable[difficulty] = {} end
  local difficultyTable = instanceTable[difficulty]
  if difficultyTable[encounterID] == nil then difficultyTable[encounterID] = {} end
  local encounterTable = difficultyTable[encounterID]
  if encounterTable[itemID] == nil then encounterTable[itemID] = {} end
  local itemTable = encounterTable[itemID]
  if itemTable[desirability] == nil then itemTable[desirability] = {} end
  local desireTable = itemTable[desirability]
  
  desireTable[characterName] = lastUpdate or time()
end

-- Send a single entry to the guild
local function sendEntry(instance, difficulty, encounter, itemID, desirability, character, lastUpdate)
  local table = {}
  addEntry(table, instance, difficulty, encounter, itemID, desirability, character, lastUpdate)
  BossList.SendUpdate(table)
end

-- Compare all the entries in table to the local data.
function BossList.MergeData(recvTable, sender)
  local latestEntry = BossList.lastUpdates[sender] or 0
  for instance, difficulties in pairs(recvTable) do           -- Instances
    for difficulty, encounters in pairs(difficulties) do      -- Difficulties
      for encounter, items in pairs(encounters) do            -- Encounters
        for item, desires in pairs(items) do                  -- Items
          for desire, characters in pairs(desires) do         -- Desires
            local lastUpdate = characters[sender]
            if lastUpdate ~= nil then
              BossList.addEntry(instance, difficulty, encounter, item, desire, sender, lastUpdate)
              
              if latestEntry < lastUpdate then
                latestEntry = lastUpdate
              end
            end
          end
        end
      end
    end
  end
  
  BossList.lastUpdates[sender] = latestEntry
end

-- Returns a table containing all entries of the user which have changed since time
function BossList.ChangesSince(time)
  local table = {}
  
  for instance, difficulties in pairs(BossList_entries) do    -- Instances
    for difficulty, encounters in pairs(difficulties) do      -- Difficulties
      for encounter, items in pairs(encounters) do            -- Encounters
        for item, desires in pairs(items) do                  -- Items
          for desire, characters in pairs(desires) do         -- Desires
            if characters[playerName] ~= nil and characters[playerName] > time then
               addEntry(table, instance, difficulty, encounter, item, desire, playerName, characters[playerName])
            end
          end
        end
      end
    end
  end
  
  return table
end

-- Deletes entries older than 3 months
local function CheckTable()
  for instance, difficulties in pairs(BossList_entries) do    -- Instances
    for difficulty, encounters in pairs(difficulties) do      -- Difficulties
      for encounter, items in pairs(encounters) do            -- Encounters
        for item, desires in pairs(items) do                  -- Items
          for desire, characters in pairs(desires) do         -- Desires
            for character, lastUpdate in pairs(characters) do -- characters
              if time() - lastUpdate > 7776000 or BossList.IsBlackListed(character) then -- Older than 3 months or blacklisted
                if desire ~= 4 and character == playerName then -- Found outdated entry of player. Create confirmation dialog
                  
                  local _, itemLink = GetItemInfo(item)
                  
                  local query = false
                  if itemLink == nil then
                    query = true
                    itemLink = "|cFFFF0000Retrieving Item Information|r"
                  end
                  local itemTexture = GetItemIcon(item)
                  local desireName = desireNames[desire]
                  local instanceName = EJ_GetInstanceInfo(instance)
                  local difficultyName = difficultyNames[difficulty]
                  local encounterName = EJ_GetEncounterInfo(encounter)
                  Dialog:Register("BossList", {
                    text = "BossList\n\nYou selected |cFFBCE000" .. desireName .. "|r for the  following item over 3 months ago:\n|T".. 
                            itemTexture .. ":32|t" .. itemLink .. "\nwhich drops in |cFF24E0D4" .. 
                            difficultyName .. "|r |cFFFFFF00" .. instanceName .. "|r from |cFF00FF00" .. encounterName ..
                            "|r\n\nDo you still need it?",
                    buttons = {
                      {
                        text = "Yes",
                        on_click = function() 
                                    local currentTime = time()
                                    BossList_entries[instance][difficulty][encounter][item][desire][character] = currentTime
                                    sendEntry(instance, difficulty, encounter, item, desire, character, currentTime)
                                    CheckTable() 
                                   end,
                      },
                      {
                        text = "No",
                        on_click = function() 
                                    BossList_entries[instance][difficulty][encounter][item][desire][character] = nil;
                                    print(string.format("BossList: Deleted entry %s %s - %s: %s (%s)",
                                          difficultyName, instanceName, encounterName, GetItemInfo(item), desireName)) 
                                    CheckTable()
                                   end,
                      }
                    },
                    on_show = function(dialog) dialog.close_button:Hide() end,
                    on_hide = function(dialog) dialog.close_button:Show() end,
                    show_while_dead = true,
                    hide_on_escape = false,
                  })
                  local dialog = Dialog:Spawn("BossList")
                  
                  dialog:SetScript("OnEnter", function(self) 
                                                GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
                                                GameTooltip:SetItemByID(item)
                                                
                                                GameTooltip:Show()
                                              end)
                  dialog:SetScript("OnLeave", function(self) 
                                                GameTooltip:Hide()
                                              end)
                  if query then
                    BossList.QueryItem(item, dialog.text)
                  end
                  return
                else 
                  BossList_entries[instance][difficulty][encounter][item][desire][character] = nil;
                end
              end
            end
            if next(characters) == nil then
              desires[desire] = nil
            end
          end
          if next(desires) == nil then
            items[item] = nil
          end
        end
        if next(items) == nil then
          encounters[encounter] =  nil
        end
      end
      if next(encounters) == nil then
        difficulties[difficulty] = nil
      end
    end
    if next(difficulties) == nil then
      BossList_entries[instance] = nil
    end
  end
end

local function BuildLastUpdateTable()
  --BossList.userEntries = {}
  BossList.lastUpdates = {}
  
  for instance, difficulties in pairs(BossList_entries) do    -- Instances
    for difficulty, encounters in pairs(difficulties) do      -- Difficulties
      for encounter, items in pairs(encounters) do            -- Encounters
        for item, desires in pairs(items) do                  -- Items
          for desire, characters in pairs(desires) do         -- Desires
            for character, lastUpdate in pairs(characters) do -- Characters
              --if character == playerName then
              --  addEntry(BossList.userEntries, instance, difficulty, encounter, item, desire, character, lastUpdate)
              --end
              if character == playerName and (BossList.lastUpdates[character] or 0) < lastUpdate then
                BossList.lastUpdates[character] = lastUpdate
              end
            end
          end
        end
      end
    end
  end
end

local function VerifyDrop(instanceID, difficulty, encounterID, itemID, noItemsIsInvalid)
  --print("Looking up: " .. instanceID .. ", " .. difficulty .. ", " .. encounterID .. ", " .. itemID)
  EJ_SelectInstance(instanceID)
  EJ_SelectEncounter(encounterID)
  
  if difficulty == 1 then difficulty = 3
  elseif difficulty == 2 then difficulty = 5
  elseif difficulty == 3 then difficulty = 14 end
  
  EJ_SetDifficulty(difficulty)
  EJ_SetLootFilter(0,0)
  
  local numItems = EJ_GetNumLoot()
  if numItems == 0 then
    if noItemsIsInvalid then
      return false
    else
      return true
    end
  end
  for itemIdx = 1, numItems do
    _, _, _, _, item, _ = EJ_GetLootInfoByIndex(itemIdx)
    
    if itemID == item then
      return true
    end
  end
  
  --print("Looking up: " .. instanceID .. ", " .. difficulty .. ", " .. encounterID .. ", " .. itemID)
  --print("Item not found")
  return false
end

local function CheckData(deleteEncounterNotFound)
  local counter = 0
  for instance, difficulties in pairs(BossList_entries) do    -- Instances
    for difficulty, encounters in pairs(difficulties) do      -- Difficulties
      for encounter, items in pairs(encounters) do            -- Encounters
        for item, _ in pairs(items) do                  -- Items
          local verification = VerifyDrop(instance, difficulty, encounter, item, deleteEncounterNotFound)
          if not(verification) then
            counter = counter + 1
            BossList_entries[instance][difficulty][encounter][item] = nil
          end
        end
      end
    end
  end

  --print(counter .. " items did not pass")
end

-- Might be used to catch changes between versions
local function CheckRevision()
  if DB.global.revision == nil or DB.global.revision < BossList.REVISION then
    --print("Clearing invalid data")
    CheckData(false)
    local AceTimer = LibStub("AceTimer-3.0")
    AceTimer:ScheduleTimer(CheckData, 10, false)
    AceTimer:ScheduleTimer(CheckData, 20, true)
    DB.global.revision = BossList.REVISION
    
  end
end

local function HideButtons()
  local items = EncounterJournal.encounter.info.lootScroll.buttons;
	local item;
	
	for i = 1,#items do
		item = items[i];
		if item.choiceButton ~= nil then
      item.choiceButton:Hide()
		end
  end
end

local function BossList_Menu_OnClick(self, desire, arg2, checked)  
  PlaySound("igMainMenuOptionCheckBoxOn")
  if not(IsInGuild()) then 
    print("BossList: You are not in a guild, you can't change your choices")
    return
  end

  dropDown.button:SetText(desireIcons[desire])
  local color = desireColors[desire]
  dropDown.button:GetFontString():SetTextColor(color.r, color.g, color.b)
  
  local instanceID = EncounterJournal.instanceID
  local item = dropDown.button:GetParent()
  local encounterID = item.encounterID
  local itemID = item.itemID
  local lastUpdate = time()
  
  BossList.lastUpdates[playerName] = lastUpdate
  BossList.addEntry(instanceID, difficulty, encounterID, itemID, desire, playerName, lastUpdate)
  sendEntry(instanceID, difficulty, encounterID, itemID, desire, playerName, lastUpdate)
end

local function BossList_Menu(frame, level, menuList)
  local desire = 4
  if dropDown ~= nil and dropDown.button ~= nil then
    local item = dropDown.button:GetParent()
    desire, _ = getEntry(BossList_entries, EncounterJournal.instanceID, difficulty, item.encounterID, item.itemID, playerName) or 4
  end
  
  local info = UIDropDownMenu_CreateInfo()
  info.func = BossList_Menu_OnClick
  for idx, desireName in pairs(desireNames) do
    info.text, info.arg1, info.checked = desireName, idx, idx == desire
    UIDropDownMenu_AddButton(info)
  end
end

local function ToggleChoiceDropDown(button)
  PlaySound("igMainMenuOptionCheckBoxOn")
  dropDown.button = button
  ToggleDropDownMenu(1, nil, dropDown, "cursor", 3, -3)
end

local function LootUpdate_Hook()
  HideButtons()
    
  if not EJ_InstanceIsRaid() or difficulty == 0 then -- Only display if the instance is a raid and difficulty is not raid finder
    return
  end
    
  local instanceID = EncounterJournal.instanceID
  local items = EncounterJournal.encounter.info.lootScroll.buttons;
	local item, index, choiceButton;
	
	for i = 1,#items do
		item = items[i];
		if item:IsVisible() then  
      if item.choiceButton == nil then -- Create a new button if needed
        choiceButton = CreateFrame("Button", "BossListChoiceButton", item, "EncounterDifficultyTemplate")
        choiceButton:Raise()
        choiceButton:SetScript("OnClick", ToggleChoiceDropDown)
        item.choiceButton = choiceButton
      else
        choiceButton = item.choiceButton
      end
      
      -- Move it in the right position depending on whether it is instance or encounter loot view
      choiceButton:ClearAllPoints()
      if EncounterJournal.encounterID == nil then -- Is nil if no encounter is selected
        choiceButton:SetPoint("RIGHT",item, 5, 18)
      else
        choiceButton:SetPoint("RIGHT",item, 5, 9)
      end
      
      -- Make button represent current choice
      local desire, _ = getEntry(BossList_entries, instanceID, difficulty, item.encounterID, item.itemID, playerName) or 4
      choiceButton:SetText(desireIcons[desire])
      local color = desireColors[desire]
      choiceButton:GetFontString():SetTextColor(color.r, color.g, color.b)

      choiceButton:Show()
		end
	end
end

local function Init()
  dropDown = CreateFrame("Frame", "BossListContextMenu", UIParent, "UIDropDownMenuTemplate")
  UIDropDownMenu_Initialize(dropDown, BossList_Menu, "MENU")
    
	hooksecurefunc(EncounterJournalNavBarHomeButton,"myclick", HideButtons);
  hooksecurefunc("EncounterJournal_LootUpdate", LootUpdate_Hook);
  hooksecurefunc(EncounterJournal.encounter.info.lootScroll, "update", LootUpdate_Hook);
end

function BossList.IsBlackListed(name)
  return tContains(DB.global.blackList, name)
end

local function EventHandler(self, event, ...)
  if event=="ADDON_LOADED" then
    local name = ...
    if name == "BossList" then   
      -- Create options
      DB = AceDB:New("BossList_DB", defaults)
      BossList.DB = DB
      
      if BossList_entries == nil then
        --print("BossList: First time loaded, initialized BossList_entries")
        BossList_entries = {}
      end
      
      BossList_entries.revision = nil;
      
      CheckRevision()
      CheckTable() -- Check for old entries
      
      BuildLastUpdateTable()
      
      BossList.Synchronize()
      
      -- Create minimap button
      Icon:Register("BossList", LDBobject, DB.char.minimap)     
      
    
    elseif name == "Blizzard_EncounterJournal" then 
      DisplayEncounter_Orig = EncounterJournal_DisplayEncounter;
      
      Init()
      frame:UnregisterEvent("ADDON_LOADED")
    end
  elseif event == "GET_ITEM_INFO_RECEIVED" then
    for itemID, textFields in pairs(queriedItems) do
      local _, itemLink = GetItemInfo(itemID)
      
      if itemLink ~= nil then
        for _, textField in pairs(textFields) do
          local text = textField:GetText()
          textField:SetText(text:gsub("|cFFFF0000Retrieving Item Information|r", itemLink))
        end
        
        queriedItems[itemID] = nil
      end
    end
  elseif event == "EJ_DIFFICULTY_UPDATE" then
    -- Normal = 1, Heroic = 2, Flexible = 3, Raid finder(and any others) = 0
    local newDifficulty = ...;
    if newDifficulty == 3 or newDifficulty == 4 then
      difficulty = 1
    elseif newDifficulty == 5 or newDifficulty == 6 then
      difficulty = 2
    elseif newDifficulty == 14 then
      difficulty = 3
    else
      difficulty = 0
    end
  end
end

-- Register chat prefixes
frame:RegisterEvent("ADDON_LOADED")
frame:RegisterEvent("GET_ITEM_INFO_RECEIVED")
frame:RegisterEvent("EJ_DIFFICULTY_UPDATE")
frame:SetScript("OnEvent", EventHandler)

optionsTable = {
  type = "group",
  args = {
    blackListHeader = {
      name = "Black list",
      order = 1,
      type = "header",
    },
    blackListDescription = {
      name = "Adding a player to the black list removes all his entries from BossList, and prevents new entries of this player from being added again. Beware, this also means you stop transmitting this player's entries to other players in the guild. As such, the black list should not be used as a filter.",
      fontSize = "medium",
      order = 2,
      type = "description",
    },    
    input = {
      name = "Add player",
      order = 3,
      width = "double",
      type = "input",
      pattern = "^[^%s%p%d%c]+$", -- No spaces, punctuation characters, digits or control characters allowed
      desc = "Enter a valid character name",
      set = function(info, string) 
              if not(tContains(DB.global.blackList, string)) then
                tinsert(DB.global.blackList, string)
                sort(DB.global.blackList)
                CheckTable()
              end
            end
    },
    blackList = {
      name = "Remove player",
      order = 4,
      type = "select",
      style = "dropdown",
      values = function() return DB.global.blackList end,
      set = function(info, val) 
              tremove(DB.global.blackList, val)
              sort(DB.global.blackList)
            end,
    },
    settingsHeader = {
      name = "Settings",
      order = 5,
      type = "header",
    },
    minimap = {
      name = "Hide minimap button",
      order = 6,
      desc = "Hides/shows the minimap button",
      width = "double",
      type = "toggle",
      set = function(info,val)
              if val then 
                Icon:Hide("BossList") 
                DB.char.minimap.hide = true
              else 
                Icon:Show("BossList")
                DB.char.minimap.hide = false
              end end,
      get = function(info) return DB.char.minimap.hide end
    },
    notNeeded = {
      name = "Hide \"Not Needed\" entries",
      order = 7,
      desc = [[Hides/shows "Not Needed" in the BossList overview
      
Warning: decreases performance when opening/filtering the overview]],
      width = "double",
      type = "toggle",
      set = function(info,val)
              if val then 
                DB.global.hideNotNeeded = true
              else 
                DB.global.hideNotNeeded = false
              end 
              if BossList.frame.visible then
                BossList.frame:Show()
              end end,              
      get = function(info) return DB.global.hideNotNeeded end
    },
    
  }
}


LibStub("AceConfig-3.0"):RegisterOptionsTable("BossList", optionsTable);
LibStub("AceConfigDialog-3.0"):AddToBlizOptions("BossList", "BossList");

-- Register slash commands
SLASH_BOSSLIST1 = "/bosslist"
SLASH_BOSSLIST2 = "/bl"
SlashCmdList["BOSSLIST"] = function(msg)
  local arg1, arg2, arg3, arg4 = strsplit(" ", msg)
  if arg1 == "resync" then
    if arg2 ~= nil then
      RequestSync(arg2)
    else
      RequestUpdater()
    end
  else
    BossList.frame:Show()
  end
end

